


void reflectXY(const Bodies &bodies, Bodies &mirrorReflection)
/** Reflection with respect to one the plane OXY
    x->x , y->y, z->-z
*/
{
    MyInterval temp;

    mirrorReflection = bodies;
    for (int i = 0; i < bodies.N - 1; i++)
    {
            temp = -bodies.z(i);
            mirrorReflection.setZ(i, temp);

    }
}





bool flatCC(const Bodies &bodies)
// returns true   the CC is planar solution (in 3D), i.e. is contained in the plane OXY (this is the only posibility to be planar)
//  uses global variables: zerosTestedFile,  undecidedFile in a non-essential way - which should be 
//   removed
{
    for (int i = 0; i < bodies.N; i++)
    {
        if (!isSingular(bodies.z(i)))
            return false;
    }

    Bodies mirrorReflection(bodies);
    Bodies hullBodies(bodies);

    reflectXY(bodies, mirrorReflection);
    intervalHullBodies(bodies, mirrorReflection, hullBodies);

    

 // the trash variables used for compatybility with the call of searchCC
    ofstream outFileTest;
	ofstream outFileFailed;  
    outFileTest.open(zerosTestedFile, ios::binary | ios::out | ios::trunc);
    outFileFailed.open(undecidedFile, ios::binary | ios::out | ios::trunc);
	// end of trash

    Count zeroFound = searchCC(hullBodies, false,outFileTest,outFileFailed); // testing stage
	                                  

    outFileTest.close();
	outFileFailed.close();

    if(isUnique(zeroFound))
       return true;
    return blowUpAndProveUniqueSolution(hullBodies);
}



MyInterval normEucl(MyInterval x, MyInterval y, MyInterval z)
/** Returns Euclidean norm of vector (x,y,z) */
{
    return sqrt((power(x, 2) + power(y, 2) + power(z, 2)));
}




//------------------------------------------------

void transposeMatrix(int size,const MyMatrix &R, MyMatrix &RT)
// on output: RT=R^T
{
    MyMatrix tempRT(size, size);
    for (int i = 0; i < size; i++)
        for (int j = 0; j < size; j++)
        {
            tempRT[i][j] = R[j][i];
        }

    for (int i = 0; i < size; i++)
        for (int j = 0; j < size; j++)
        {
            RT[i][j] = tempRT[i][j];
        }
}

void writeVec(MyInterval x, MyInterval y, MyInterval z)
{
    cout << "(" << x << ", " << y << ", " << z << ") ";
}

MyInterval distFromLine(MyInterval xk, MyInterval yk, MyInterval zk, MyInterval dirX, MyInterval dirY, MyInterval dirZ)
//  returns: the distance of the point (xk,yk,zk) from the line t*(dirX,dirY,dirZ)
//      it is assumed that |(dirX,dirY,dirZ)|=1
{
    MyInterval sc = scalarProduct(xk, yk, zk, dirX, dirY, dirZ);
    return normEucl(xk - sc * dirX, yk - sc * dirY, zk - sc * dirZ);
}

void vectProduct(MyInterval dirEXx, MyInterval dirEXy, MyInterval dirEXz,
                 MyInterval dirEYx, MyInterval dirEYy, MyInterval dirEYz,
                 MyInterval &dirEZx, MyInterval &dirEZy, MyInterval &dirEZz)
// On output:   MyInterval (dirEZx,dirEZy,dirEZz)=(dirEXx,dirEXy,dirEXz) x (dirEYx,dirEYy,dirEYz)
				 
{
    dirEZx = dirEXy * dirEYz - dirEXz * dirEYy;
    dirEZy = dirEXz * dirEYx - dirEXx * dirEYz;
    dirEZz = dirEXx * dirEYy - dirEXy * dirEYx;
}

void candidatesForB(const Bodies &bodies, int bodyA, MyInterval dirx, MyInterval diry, MyInterval dirz,
                    vector<int> &possibleB)
// in the construction of possible ortogonal transformation we first identify bodyA which will be image of 	q_{n-2}
//    this gives us the direction (dirx,diry, dirz) - the normalized direction of bodyA - an image of the axis OX	
//  now the bodyB, the image of q_0, is found as the point which is furtherest from this line			
/*
 * possibleB[] contains indicies of bodies which might be images of q_0 
 */
{
    int N = bodies.N;
    possibleB.clear();
    MyInterval absY0 = abs(bodies.y(0));
    //cout << "absY0" << absY0 << endl;
    

    for (int k = 0; k < N; k++)
    {
        if (k == bodyA)
            continue;
        MyInterval xk = bodies.x(k);
        MyInterval yk = bodies.y(k);
        MyInterval zk = bodies.z(k);
        MyInterval tk = distFromLine(xk, yk, zk, dirx, diry, dirz);
       

        if (!emptyIntersection(tk, absY0))
        {
            possibleB.push_back(k);
        }
    }
}

void writePermCycles(vector<vector<int>> &cycles, int noC)
//  we write to cout the decomposion of permutation into cycles
{
    for (int lc = 0; lc < noC; lc++)
    {
        const int lenCyc = cycles[lc].size();
        cout << "(";
        for (int c = 0; c < lenCyc; c++)
        {
            if (c != 0)
                cout << ",";
            cout << cycles[lc][c];
        }
        cout << ")";
    }
}

bool minusIdentity(const MyMatrix &R, int size)
// check if R contains -Id
{
    for (int i = 0; i < size; i++)
        for (int j = 0; j < size; j++)
        {
            if (i == j && (!R[i][j].contains(-1)))
                return false;

            if (!R[i][j].contains(0))
                return false;
        }
    return true;
}

MyInterval trace(const MyMatrix &R, int dim)
// returns trace(R)
{
    MyInterval tr = 0;
    for (int i = 0; i < dim; i++)
    {
        tr += R[i][i];
    }
    return tr;
}

void writesymmetryFromPerm(int permutation[], int N, vector<vector<int>> &cycles, int noC, bool sg, MyMatrix R, int size)
// the function writes to cout the data about symmetry
{
    // sg --- sign of det of matrix R:
    // sg = TRUE <==> det(R) = 1 <==> R+
    // sg = FALSE <==> det(R) = -1 <==> R-
    // noC -- number of cycles in vector 'cycles'
    int oneElementC = 0;  // the noumber of 1-element cycles
    int twoElementC = 0;  // the noumber of 2-element cycles
    int manyElementC = 0; // the noumber of many (>2)-element cycles
    int longestC = 1; // length of the longest cycle

    for (int lc = 0; lc < noC; lc++)
    {
        int lenCyc = cycles[lc].size();
        if (lenCyc > longestC)
            longestC = lenCyc;

        if (lenCyc == 1)
        {
            oneElementC++; 
        }
        else if (lenCyc == 2)
        {
            twoElementC++; 
        }
        else
        {
            manyElementC++; 
        }
    }

    if (sg) // R+
    {
        bool firstO = true;
        cout << "det(R) = 1" << endl;
        for (int lc = 0; lc < noC; lc++)
        {
            int lenCyc = cycles[lc].size();
            if (lenCyc == 1)
            {
                if (firstO)
                {
                    cout << "Rotation axis: 0";
                    firstO = false;
                }
                cout << "-q" << cycles[lc][0];
            }
        }

        cout << endl
             << "Rotation angle = 2*Pi/" << longestC << endl
             << endl;
    }
    else // R-
    {
        cout << "det(R) = -1" << endl;

        MyInterval tr = trace(R, 3);
        //cout << "trace R = " << tr << endl;
        if (manyElementC == 0) // either (-Id) or reflection with respect to some plane
        {
            if (twoElementC == 0)
                cout << "Reflection with respect to the plane Z = 0" << endl;
            else
            {
                if (tr < 1)
                {
                    cout << "-Identity (-Id)" << endl;
                }
                else if (tr > -3)
                {
                    int qi = 0;
                    int qj = 0;
                    for (int lc = 0; lc < noC; lc++)
                    {
                        int lenCyc = cycles[lc].size();
                        if (lenCyc == 2)
                        {
                            qi = cycles[lc][0];
                            qj = cycles[lc][1];
                        }
                    }
                    // reflection with respect to the plane qj - qi
                    cout << "Reflection with respect to the bisecting plane of the segment "
                         << "[q" << qj << ", q" << qi << "]" << endl;
                }
                else // it should never happen!!!
                {
                    cout << "We cannot establish a kind of symmetry!" << endl;
                }
            }
        }
        else
        {
            MyInterval cosPhi = (tr + 1) / 2.0;
            MyInterval posCos1 = cos(2 * piConst / longestC);
            MyInterval posCos2 = cos(4 * piConst / longestC);
            //cout << "cosPhi = " << cosPhi << ", posCos1 = " << posCos1 << ", posCos2 = " << posCos2 << endl;
            if (!emptyIntersection(cosPhi, posCos1) && emptyIntersection(cosPhi, posCos2))
            {
                cout << "Rotation angle = 2*Pi/" << longestC << endl
                     << endl;
            }
            else if (emptyIntersection(cosPhi, posCos1) && !emptyIntersection(cosPhi, posCos2))
            {
                cout << "Rotation angle = Pi/" << longestC << endl
                     << endl;
            }
            else if (!emptyIntersection(cosPhi, posCos1) && !emptyIntersection(cosPhi, posCos2))
            { // it should never happen!!!
                cout << "Rotation angle = 2*Pi/" << longestC << " or Pi/" << longestC << endl
                     << endl;
            }
            else // it should never happen!!!
            {
                cout << "An error: in computation of rotation angle!" << endl;
            }
        }
    }
}


void permCycles(int permutation[], int N, vector<vector<int>> &cycles, int &noC)
// decomposition of permutation into cycles
// On input:
//    N - the number of elements in permutation
//    permutation[] - the permutation to be decomposed into cycles
// On output:
// cycles[k][0] -- cykl zaczynajacy sie od i
// noC -- liczba cykli --- wspolrzedne w wektorze cycles sa od 0 do (noC-1)
{

    int uI[N];  // uI[i]==0 for elements which are not yet in the cycle decomposition
    for (int i = 0; i < N; i++)
        uI[i] = 0;

    noC = 0;
    for (int i = 0; i < N; i++)
    {
        if (uI[i] != 0)  //  the element i is already in some cycle
            continue;

        uI[i] = 1;
        vector<int> row;  // Create an empty row
        row.push_back(i); // Add first element of cycle to the row
        int j, k;
		// the construction of the cycle containing i-th element
        k = i;
        do
        {
            j = permutation[k];
            if (uI[j] == 0) // j-the element was not visited yet, hence j!=i and we add it to the cycle
            {
                row.push_back(j); // Add another element of cycles to the row
                uI[j] = 1;
                k = j;
            }
        } while (j != i);
        cycles.push_back(row); // Add the row to the main vector

        noC++;
    }
}
// */

void writeDataOnSymmetry(int permutation[], int N,const MyMatrix &RT, bool sg)
// writes to cout the data about symmetry
// sg --- sign of det of matrix R:
    // sg = TRUE <==> det(R) = 1 <==> R+
    // sg = FALSE <==> det(R) = -1 <==> R-
{
    cout << endl;
    cout << "--------------------------------------------------------------------------------" << endl;
    cout << "Permutation:";
    for (int i = 0; i < N; i++)
        cout << " " << permutation[i];
    vector<vector<int>> cycles;
    int noC;
    permCycles(permutation, N, cycles, noC);
    cout << ", cycles: ";
    writePermCycles(cycles, noC);
    cout << endl;
    cout << "Orthogonal transformation: " << RT << endl;
    writesymmetryFromPerm(permutation, N, cycles, noC, sg, RT, 3);
}

bool proveSymmetryWriteToReportFile(const Bodies &bodies, int permutation[], MyMatrix &RT, bool sg)
//  the function proves that (permutation,RT) is a symmetry and writes its description to the report file (cout)
//  bodies - is a box containing a unique CC
//  the symmetry is given as a pair (permutation[],RT) 
// // sg --- sign of det of matrix R:
    // sg = TRUE <==> det(R) = 1 <==> R+
    // sg = FALSE <==> det(R) = -1 <==> R-
//  Return "true" - is the symmetry has been proved
//         "false" - otherwise      


{
    
    int N = bodies.N;
    int invPerm[N];
    for (int i = 0; i < N; i++)
        invPerm[permutation[i]] = i;
		
	

    Bodies symBodies(bodies);
    // below we construct symBodies --- a set which contains interval hull of bodies and its symmetric image
    MyInterval RTxi, xj;
    MyInterval RTyi, yj;
    MyInterval RTzi, zj;
    MyVector qi(3);

    for (int i = 0; i < N; i++)
    {
        if (permutation[i] == i)
        {
            continue;
        }
        else
        {
            int j = invPerm[i];
            if (j == N - 1)
                continue;
            xj = bodies.x(j);
            yj = bodies.y(j);
            zj = bodies.z(j);

            qi[0] = bodies.x(i);
            qi[1] = bodies.y(i);
            qi[2] = bodies.z(i);

            MyVector RTqi = RT * qi;
            RTxi = RTqi[0];
            RTyi = RTqi[1];
            RTzi = RTqi[2];

            symBodies.setX(j, intervalHull(RTxi, xj));
            symBodies.setY(j, intervalHull(RTyi, yj));
            symBodies.setZ(j, intervalHull(RTzi, zj));
        }
    }

     // the trash variables used for compatybility with the call of searchCC
    ofstream outFileTest;
	ofstream outFileFailed;  
    outFileTest.open(zerosTestedFile, ios::binary | ios::out | ios::trunc);
    outFileFailed.open(undecidedFile, ios::binary | ios::out | ios::trunc);
	// end of trash

    bool retval=false;
    Count zeroFound = searchCC(symBodies, false,outFileTest,outFileFailed);

    outFileTest.close();
	outFileFailed.close();

    if (zeroFound.maybeZeros == 0)
    {
        if (zeroFound.zeros == 1)
        {
            retval=true;
        }
        else if (zeroFound.zeros > 1)
        {
            cout << "Couldn't establish uniqueness!" << endl;
            return false;
        }
    }

    // the search is inconclusive likely because the set is to small for our accuracy
    // we try to blow up this set
	if(!retval)
	{
      if(blowUpAndProveUniqueSolution(symBodies)){
	     retval=true;
      }
	  else
   	  {
        cout << "undecided case!" << endl;
        return false;
      }
	}  
	
	
	
	if(retval)
	   writeDataOnSymmetry(permutation, N, RT, sg);
    return retval;
} 

bool establishPermutation(const Bodies &bodies, int permutation[], MyMatrix &R)
// for a given ortogonal matrix R the function tries to find a permutation[]
// such that (permutation,R) will be the symmetry of CC contained in bodies
// bodies - is a box containing a unique CC
{

    int N = bodies.N;
    bool found;
    for (int i = 0; i < N; i++)
    {
        found = false;
    /*    if (permutation[i] != -1)
        {
            found = true;
            continue;
        }
*/
        MyVector qi(3);
        qi[0] = bodies.x(i);
        qi[1] = bodies.y(i);
        qi[2] = bodies.z(i);

        MyVector Rqi(3);
        Rqi = R * qi;
        //        cout << "Rq" << i << " = " << Rqi << endl;

        // now we look for pair for qi
        for (int j = 0; j < N; j++)
        {
            MyInterval xj = bodies.x(j);
            MyInterval yj = bodies.y(j);
            MyInterval zj = bodies.z(j);

            if (!emptyIntersection(Rqi[0], xj) &&
                !emptyIntersection(Rqi[1], yj) &&
                !emptyIntersection(Rqi[2], zj)) // qj cap Rqi \neq 0
            {
                //              cout << "dla " << i << "znalazl " << j << endl;
                permutation[i] = j;
                found = true;
                break;
            }
        }
        if (!found) // there is no pair for qi body --- this is not proper permutation
        {
            //cout << "there is no pair for q" << i << " body --- this is not proper permutation" << endl;
            return false;
        }
    }

   
    return true; 
}

int FindAndProveOrthogonalSymmetriesWithBodyA(const Bodies &bodies, int bodyA)
// the function constructs all ortogonal symmetries such that N-2 ->  bodyA
//       and does the rigorous verification 
//   returns number of ortogonal symmetries different from identity
//
//  in the construction of possible ortogonal transformation we first identify bodyA which will be image of 	q_{n-2}
//    this gives us the direction (dirx,diry, dirz) - the normalized direction of bodyA - an image of the axis OX	
//  now the bodyB, the image of q_0, is found as the point which is furtherest from this line	
/**
 *   permutation[N-2] = bodyA;
 *   permutation[0] = bodyB;
     this fixes the image of the plane OXY
	 and then it remains the decide what will be the image of OZ-axis , we have two possibilities the orientation preserving and non-preserving
 *
 *
 */
{

    int numOfSym = 0;
    int N = bodies.N;
   
	

    MyInterval xA = bodies.x(bodyA);
    MyInterval yA = bodies.y(bodyA);
    MyInterval zA = bodies.z(bodyA);
    // ex = (dirEXx, dirEXy, dirEXz) --- image of coordinates axis  OX by  R
    MyInterval distxA = normEucl(xA, yA, zA);
    MyInterval dirEXx = xA / distxA;
    MyInterval dirEXy = yA / distxA;
    MyInterval dirEXz = zA / distxA;
   
    
    vector<int> possibleB;

    candidatesForB(bodies, bodyA, dirEXx, dirEXy, dirEXz, possibleB);

  
    bool RminusOnly;
    for (auto &&bodyB : possibleB)
    {
	 
        RminusOnly = false;
        if (bodyA == N - 2 && bodyB == 0) // R would be identity or reflection z --> -z
        {
            RminusOnly = true;  // the orientation preserving is the identity
        }
       
        
        // we construct ey = (dirEYx, dirEYy, dirEYz)
        MyInterval xB = bodies.x(bodyB);
        MyInterval yB = bodies.y(bodyB);
        MyInterval zB = bodies.z(bodyB);

        MyInterval sc = scalarProduct(xB, yB, zB, dirEXx, dirEXy, dirEXz);
        MyInterval tildeX = xB - sc * dirEXx;
        MyInterval tildeY = yB - sc * dirEXy;
        MyInterval tildeZ = zB - sc * dirEXz;
        MyInterval normTildeEy = normEucl(tildeX, tildeY, tildeZ);
        //cout << "normTildeEy = " << normTildeEy << endl;

        if (isSingular(normTildeEy))
        {
            cout << "Cannot normalize ey vector --- likely collinear vectors! " << endl;
            return false;
        }
        MyInterval dirEYx = (tildeX) / normTildeEy;
        MyInterval dirEYy = (tildeY) / normTildeEy;
        MyInterval dirEYz = (tildeZ) / normTildeEy;
      

        // ez = ex X ey
        MyInterval dirEZx;
        MyInterval dirEZy;
        MyInterval dirEZz;
        vectProduct(dirEXx, dirEXy, dirEXz, dirEYx, dirEYy, dirEYz, dirEZx, dirEZy, dirEZz);
        MyInterval normEz = normEucl(dirEZx, dirEZy, dirEZz);
        if (isSingular(normEz))
        {
            cout << "Cannont normalize ez vector! " << endl;
            return false;
        }

      
        /** matrix R  = [ex| ey| +- ez]
    * tantatively maps q_{N-2} onto q_{bodyA}
    * q_0 onto q_{bodyB}
    * RT is inverse of R
    */

        MyMatrix R(3, 3);
        ///*
        R[0][0] = dirEXx;
        R[0][1] = dirEYx;
        R[0][2] = dirEZx;
        R[1][0] = dirEXy;
        R[1][1] = dirEYy;
        R[1][2] = dirEZy;
        R[2][0] = dirEXz;
        R[2][1] = dirEYz;
        R[2][2] = dirEZz;
        //*/
        MyMatrix RT(3, 3);
        transposeMatrix(3, R, RT);

        MyMatrix Rminus(3, 3);
        Rminus = R;
        Rminus[0][2] = -dirEZx;
        Rminus[1][2] = -dirEZy;
        Rminus[2][2] = -dirEZz;
        MyMatrix RTminus(3, 3);
        transposeMatrix(3, Rminus, RTminus);
       
        // up to now we construced R and R^{-1}=RT such that
		//  there is possibility that N-2 -> bodyA
		//                             0 -> bodyB
		
		
        int *permutation=new int[N];
        if (!RminusOnly)
        {
            if (establishPermutation(bodies, permutation, R))
			{
                if (proveSymmetryWriteToReportFile(bodies, permutation, RT, true))  // true - means orientation preserving
                    numOfSym++;
			}		
        }

       
        if (establishPermutation(bodies, permutation, Rminus))
		{
		    
            if (proveSymmetryWriteToReportFile(bodies, permutation, RTminus, false)) // false - means orientation reversing
                numOfSym++;
		}		
        delete [] permutation;
    } // for --- all possible coorB
    return numOfSym;
}

int FindAndProveAllOrthogonalSymmetries(const Bodies &bodies)
// finds and proves all orthogonal symmetries different from identity
// returns number of  orthogonal symmetries different from identity
{
  
    int N = bodies.N;

    // the vector containing bodies equidistant from 0
    // vec[i][0] = i, vec[i][j] = k <=> bodies i and k have the same distance from 0
    vector<vector<int>> vec;
    findGroups(bodies, vec);

    int numEqualDist = vec[N - 2].size();

   
    int sym = 0;
    for (int j = 0; j < numEqualDist; j++) // loops through each body potentially equidistant with q_{n-2} 
    {
        sym += FindAndProveOrthogonalSymmetriesWithBodyA(bodies, vec[N-2][j]);
    }
    return sym;
}


void writeDiffCCandTestSymmetry(int ccNo, vector<Bodies> &ccSolutions, vector<int> &markedSolutions)
//  ccNo - the number of entries in ccSolutions, markedSolutions
//      ccSolution[i] is treated as a representant of CC iff markedSolutions[i]==0
//   the function writes to the standard output: different CCs together with the information of their symmetries
//                  for !equalmasses it makes sense to test only the collinearity and flatness
//                   otherwise we check all possible orthogonal transformations
//  the function uses global variables: mathFile,mathListFile
{
   
   
    int symColCounter = 0;
    int symFlatCounter = 0;
    //int symLineCounter = 0;

    bool doNotTestFlatSym = true; // if TRUE, program does not test symmetries for flat CC

    cout << endl;
    cout << "Different CC:" << endl;
    int realCC = 0;
    //int counter = 0; // the flag of symmetries by lines:
    // for each body is 1 if there is a symmetry or 0 fi there is no

    for (int i = 0; i < ccNo; i++)
    {
        if (markedSolutions[i] == 0)
        {
            realCC++;
            mathFile << "---------------------" << endl
                     << endl;
            mathFile << "position " << i << endl;
            ccSolutions[i].writeCCformatMath(mathFile,mathListFile);

            cout << "======================================================================================================" << endl;
            cout << "position " << i << endl;
            cout << "======================================================================================================" << endl;
            ccSolutions[i].printData(true); // writes distances from 0
            cout << flush;

            writePotAndInertia(ccSolutions[i],false);
            bool symCol = false;
            if (colinearCC(ccSolutions[i]))
            {
                symColCounter++;
                symCol = true;
                cout << "collinear solution no " << symColCounter << endl;
            }
            bool symFlat = false;
            if (flatCC(ccSolutions[i]))
            {
                symFlatCounter++;
                symFlat = true;
                cout << "planar solution no " << symFlatCounter << endl;
                if (symCol || (symFlat && doNotTestFlatSym))
                    continue; 
            }
            else
            {
                MyInterval det;
                if (!detCondForReducedSys(ccSolutions[i], det))
                {
                    cout << "Solution of reduced set of equations, but we cannot claim that this is CC!" << endl;
                    cout << "Determinant is " << det << "singular!" << endl;
                }
            }

            if (equalMasses)
            {
                int symCount;
                cout << endl;
                symCount = FindAndProveAllOrthogonalSymmetries(ccSolutions[i]);
                cout << endl
                     << "Number of non-trivial symmetries = " << symCount << endl;
            }
        }
    } // for
    cout << endl;
    cout << "Number of different cc = " << realCC << endl;
}
