#include <iostream>

#include "defo.h"


#include "cc-nb.h"


#include "bodies.h"
#include "test-and-help-functions.h"
#include "solutions-tests.h"
#include "MyIntervalFunc.h"
#include "searchalg.h"


void writePermutation(ostream& fhandle ,int numBodies, int permutation[])
{
  for(int i=0;i<numBodies;i++)
    fhandle << "  " << permutation[i];
  fhandle << endl;	
}


bool permutationOK(int numBodies, int permutation[])
//  return: true - if   permutation[n-2] =n-2   // because we drop y_{n-2} and z_{n-2}
//                  and in the spatial case permutation[0]=0  //  we drop also x_0
{
  bool retval=true;
  if(permutation[numBodies-2]!=numBodies-2)
    return false;
  #ifdef SPATIAL
  if(permutation[0]!=0)
    return false;
  #endif	
  return retval;
}


bool theSameSolution(const Bodies &sol1,const  Bodies &sol2, int permutation[],const char* zerosTestedFileName,const char* undecidedFileName)
//  function assumes that sol1 and sol2 contain a unique CC (the solution of the reduced system)
//    permutation - gives the permutation of bodies to be applied to <sol2>
//      in permutation(sol2), q_i=q_{permutation[i]}
//  returns true if sol1==permutation(sol2)
//  the permutation has to satisfy :  permutation[n-2] =n-2   // because we drop y_{n-2} and z_{n-2}
//                  and in the spatial case permutation[0]=0  //  we drop also x_0
//  the last two parameters are non-essential - are here for some compatibility , we call searchCC which writes to some files 
//               we do not want it to override something important
//  zerosTestedFileName - the name for the auxliary binary file, where the result of local search algorithm will be written, this can be used for some
//                     diagnostic purposes
//  undecidedFileName -  the name for the auxliary binary file, where the result of local search algorithm will be written, when undecided box arises, which
//                       is very unlikely
{

     int N = sol1.N;
	 
    if(!permutationOK(N,permutation))
	{
	  cout << "Problematic permutation in function theSameSolution " << endl;
	  writePermutation(cout,N,permutation);
	  cout << "---------------" << endl;
	}

   

    Bodies permSol(sol2); // solution sol2 permuted according to permutation[]
	// we permute bodies in <sol2> and check non-empty intersection with <sol1>
    for (int i = 0; i < N; i++)
    {
        MyInterval x1 = sol1.x(i), x2 = sol2.x(permutation[i]);
        if (emptyIntersection(x1, x2))
            return false; // different solutions
        if (i < N - 1)
            permSol.setX(i, x2);

        
        MyInterval y1 = sol1.y(i), y2 = sol2.y(permutation[i]);
        if (emptyIntersection(y1, y2))
              return false; // different solutions
	      
        if ((i < N - 1) && (i!=N-2)) // because y_{n-2}=0
             permSol.setY(i, y2);
			
			
		#ifdef SPATIAL
        MyInterval z1 = sol1.z(i), z2 = sol2.z(permutation[i]);
        if (emptyIntersection(z1, z2))
            return false; // different solutions
        if ((i < N - 1)  && (i!=0) && (i!=N-2))   // because z_0=z_{n-2}=0
            permSol.setZ(i, z2);
        #endif	
    }
    // non-empty intersection

    Bodies unionBodies(sol1);
    intervalHullBodies(sol1, permSol, unionBodies);

   // we define some trash variables for compatybility with the call of searchCC
    ofstream outFileTest;
	ofstream outFileFailed;  
    outFileTest.open(zerosTestedFileName, ios::binary | ios::out | ios::trunc);
    outFileFailed.open(undecidedFileName, ios::binary | ios::out | ios::trunc);
	// trash-end

    bool searchStage=false;
    Count cZeros = searchCC(unionBodies, searchStage,outFileTest,outFileFailed);
	outFileTest.close();
	outFileFailed.close();
	
    if (cZeros.zeros == 1)
        return true; // sol1 and permutation(sol2) are the same solution
    else
    {
        return blowUpAndProveUniqueSolution(unionBodies);
    }
}




bool nextPerm(int p[], int k)
//  PZ: zupelnie nie wiem o co w tym chodzi
{
    bool found = false;
    int j;
    int i = k - 1;
    do
    {
        if (p[i] > p[i - 1])
            found = true;
        else
            i--;
    } while (i > 0 && !found);

    if (!found && i == 0)
        return false; // there is no next permutation

    // we look for the smallest number among p[i], ..., p[N-1], greater than p[i-1]
    for (j = k - 1; j > i - 1; j--)
        if (p[j] > p[i - 1])
            break;

    std::swap(p[i - 1], p[j]);

    for (j = k - 1; i < j; i++, j--)
        std::swap(p[i], p[j]);
    return true;
}


bool nextPermutationToCheck(int numBodies, int perm[])
// PZ: nie wiem o co w tym chodzi
{
    bool found;
    int p[numBodies - 1];

    for (int i = 0; i < numBodies - 2; i++)
        p[i] = perm[i];
    p[numBodies - 2] = perm[numBodies - 1];
    found = nextPerm(p, numBodies - 1);
    for (int i = 0; i < numBodies - 2; i++)
        perm[i] = p[i];
    perm[numBodies - 1] = p[numBodies - 2];

    return found;
}







void identifyDifferentSolutions(bool equalmasses, int &ccNo, vector<Bodies> &ccSolutions,
                    vector<int> &markedSolutions, int &uniqueSol,const char* zerosTestedFileName, const char* undecidedFileName)
// ccNo -- the number of read solutions (the size of ccSolutions and markedSolutions arrays)
//  if !equalmasses then we just compare the solutions  (using Krawczyk operator)
//      otherwise we allow for permutations of bodies
// On output:
//  markedSolutions[i] =1 - then it is not counted as different solution in ccSolutions
//  markedSolutions[i] =0 - then it selected as a representative of unique solution (in the case of equalmasses we check again whether the ordering
//                                                           of bodies   is ok) 
//   uniqueSol -- the number of different (unique) solutions in ccSolutions
//  zerosTestedFileName - the name for the auxliary binary file, where the result of local search algorithm will be written, this can be used for some
//                     diagnostic purposes
//  undecidedFileName -  the name for the auxliary binary file, where the result of local search algorithm will be written, when undecided box arises, which
//                       is very unlikely

{
 
    markedSolutions.clear();
 
    uniqueSol = 0;

    for (int i = 0; i < ccNo; i++)
        markedSolutions.push_back(0);

    for (int i = 0; i < ccNo; i++)
    {
        Bodies sol1(ccSolutions[i]);
        for (int j = i + 1; j < ccNo; j++)
        {
            if (markedSolutions[j] == 1)
                continue;

            // initialize permutations[] to identity
			int numBodies=sol1.N;
            int permutations[numBodies];
            for (int k = 0; k < numBodies; k++)
                permutations[k] = k;

            Bodies sol2(ccSolutions[j]);
            for (;;)
            {
                if (theSameSolution(sol1, sol2, permutations,zerosTestedFileName,undecidedFileName))
                {
                    markedSolutions[j] = 1;
                    break;
                } // sol1 and sol2 are the same solution
                if (!equalmasses)
                    break; // permutations make sense only for equal masses

                if (!nextPermutationToCheck(numBodies, permutations))  // OBECNIE ROBI ZA DUZO PERMUTACJI W PRZYPADKU 3D
                    break; // there is no next permutation
            }
        }
    }

    for (int i = 0; i < ccNo; i++)
        if (markedSolutions[i] == 0)
        {
            if (!distanceTest(ccSolutions[i]))  // this might be usefull in the case of equalmasses, we remove the solutions not satisfying some constrains
			                                  // on the variables 
            {
                markedSolutions[i] = 1;
            }
            else
                uniqueSol++;
        }
}

